bl_info = {
    "name": "DECALmachine",
    "author": "MACHIN3, AR, MX",
    "version": (1, 9, 1),
    "blender": (2, 80, 0),
    "location": "Pie Menu: D key, MACHIN3 N Panel",
    "revision": "763ba532f74ace35e0bae1e827cd0cdc629f5bb4",
    "description": "A complete mesh-decal pipeline: project decals on curved surfaces, create decal panel strips, design your own custom decals, export decal assets by baking them.",
    "warning": "",
    "wiki_url": "https://decal.machin3.io",
    "tracker_url": "https://machin3.io/DECALmachine/docs/faq/#get-support",
    "category": "Mesh"}


def reload_modules(name):
    """
    This makes sure all modules are reloaded from new files, when the addon is removed and a new version is installed in the same session,
    or when Blender's 'Reload Scripts' operator is run manually.
    It's important, that utils modules are reloaded first, as operators and menus import from them
    """

    import os
    import importlib

    # first fetch and reload all utils modules
    utils_modules = sorted([name[:-3] for name in os.listdir(os.path.join(__path__[0], "utils")) if name.endswith('.py')])

    for module in utils_modules:
        impline = "from . utils import %s" % (module)

        print("reloading %s" % (".".join([name] + ['utils'] + [module])))

        exec(impline)
        importlib.reload(eval(module))

    # then update the classes and keys dicts
    from . import dicts
    importlib.reload(dicts)

    # and based on that, reload the modules containing operator and menu classes
    modules = []

    for label in dicts.classes:
        entries = dicts.classes[label]
        for entry in entries:
            path = entry[0].split('.')
            module = path.pop(-1)

            if (path, module) not in modules:
                modules.append((path, module))

    for path, module in modules:
        if path:
            impline = "from . %s import %s" % (".".join(path), module)
        else:
            impline = "from . import %s" % (module)

        print("reloading %s" % (".".join([name] + path + [module])))

        exec(impline)
        importlib.reload(eval(module))


if 'bpy' in locals():
    reload_modules(bl_info['name'])


import bpy
from bpy.props import StringProperty, BoolProperty, EnumProperty, PointerProperty, CollectionProperty, IntVectorProperty
from . properties import DecalSceneProperties, DecalObjectProperties, DecalMaterialProperties, DecalImageProperties, DecalCollectionProperties, ExcludeCollection
from . utils.registration import get_core, get_menus, get_tools, get_prefs, register_classes, unregister_classes, register_keymaps, unregister_keymaps
from . utils.registration import register_decals, unregister_decals, register_instant_decals, unregister_instant_decals, register_lockedlib, unregister_lockedlib
from . utils.registration import register_icons, unregister_icons, register_infotextures, unregister_infotextures, register_infofonts, unregister_infofonts, register_panels, unregister_panels
from . utils.registration import add_context_menu, remove_context_menu
from . utils.system import get_PIL_image_module_path
from . utils.update import update_check
from . handlers import update_panels, update_match_material_enum, update_infodecal_sources, update_userdecallibs_enum


def register():
    global classes, keymaps, icons

    # CORE

    core_classes = register_classes(get_core())


    # PROPERTIES

    bpy.types.Scene.DM = PointerProperty(type=DecalSceneProperties)
    bpy.types.Object.DM = PointerProperty(type=DecalObjectProperties)
    bpy.types.Material.DM = PointerProperty(type=DecalMaterialProperties)
    bpy.types.Image.DM = PointerProperty(type=DecalImageProperties)
    bpy.types.Collection.DM = PointerProperty(type=DecalCollectionProperties)

    bpy.types.WindowManager.matchmaterial = EnumProperty(name="Materials, that can be matched", items=[("None", "None", "", 0, 0),
                                                                                                       ("Default", "Default", "", 0, 1)])

    bpy.types.WindowManager.collectinfotextures = BoolProperty()
    bpy.types.WindowManager.excludeimages = CollectionProperty(type=ExcludeCollection)

    bpy.types.WindowManager.collectinfofonts = BoolProperty()
    bpy.types.WindowManager.excludefonts = CollectionProperty(type=ExcludeCollection)

    bpy.types.WindowManager.decal_mousepos = IntVectorProperty(name="Mouse Position for Decal Insertion", size=2)


    # DECALS

    decals = register_decals()
    register_instant_decals()
    register_lockedlib()


    # INFO DECAL SOURCES

    register_infotextures()
    register_infofonts()


    # TOOLS, MENUS, KEYMAPS

    menu_classlists, menu_keylists = get_menus()
    tool_classlists, tool_keylists = get_tools()

    classes = register_classes(menu_classlists + tool_classlists) + core_classes
    keymaps = register_keymaps(menu_keylists + tool_keylists)

    add_context_menu()

    register_panels(getall=True)


    # ICONS

    icons = register_icons()


    # HANDLERS

    bpy.app.handlers.load_post.append(update_panels)

    bpy.app.handlers.undo_post.append(update_match_material_enum)
    bpy.app.handlers.redo_post.append(update_match_material_enum)
    bpy.app.handlers.load_post.append(update_match_material_enum)

    bpy.app.handlers.load_post.append(update_userdecallibs_enum)

    bpy.app.handlers.depsgraph_update_post.append(update_infodecal_sources)


    # CHECK PIL and set preferences.pil prop accordingly

    try:
        import PIL
        from PIL import Image

        get_prefs().pil = True
        get_prefs().pilrestart = False
        path = get_PIL_image_module_path(Image)
    except:
        get_prefs().pil = False
        get_prefs().pilrestart = False
        path = ''


    # REGISTRATION TERMINAL OUTPUT

    print("Registered %s %s with %d decal libraries. %s" % (bl_info["name"], ".".join([str(i) for i in bl_info['version']]), len(decals), "PIL %s %s" % (PIL.__version__, "Image Module: %s" % path) if get_prefs().pil else "PIL is not installed."))

    for lib in decals:
        print(" » decal library: %s" % (lib))


    # INITIALIZE
    get_prefs().update_available = False


    # CHECK for new versions

    update_check()


def unregister():
    global classes, keymaps, icons

    # HANDLERS

    bpy.app.handlers.load_post.remove(update_panels)

    bpy.app.handlers.undo_post.remove(update_match_material_enum)
    bpy.app.handlers.redo_post.remove(update_match_material_enum)
    bpy.app.handlers.load_post.remove(update_match_material_enum)

    bpy.app.handlers.load_post.remove(update_userdecallibs_enum)

    bpy.app.handlers.depsgraph_update_post.remove(update_infodecal_sources)


    # DECALS

    unregister_decals()
    unregister_instant_decals()
    unregister_lockedlib()


    # INFO DECAL SOURCES

    unregister_infotextures()
    unregister_infofonts()


    # KEYMAPS

    unregister_keymaps(keymaps)


    # ICONS

    unregister_icons(icons)


    # PANELS

    unregister_panels()


    # PROPERTIES

    del bpy.types.Scene.DM
    del bpy.types.Object.DM
    del bpy.types.Material.DM
    del bpy.types.Image.DM
    del bpy.types.Collection.DM

    del bpy.types.WindowManager.matchmaterial

    del bpy.types.Scene.userdecallibs
    del bpy.types.WindowManager.newdecalidx
    del bpy.types.WindowManager.decaluuids
    del bpy.types.WindowManager.paneldecals
    del bpy.types.WindowManager.instantdecaluuids

    del bpy.types.WindowManager.collectinfotextures
    del bpy.types.WindowManager.updateinfotextures
    del bpy.types.WindowManager.excludeimages

    del bpy.types.WindowManager.collectinfofonts
    del bpy.types.WindowManager.updateinfofonts
    del bpy.types.WindowManager.excludefonts

    del bpy.types.WindowManager.decal_mousepos


    # TOOLS, MENUS, CORE

    remove_context_menu

    unregister_classes(classes)

    print("Unregistered %s %s" % (bl_info["name"], ".".join([str(i) for i in bl_info['version']])))
