import bpy
import os
from datetime import datetime
from .. utils.registration import get_prefs, get_path
from .. utils.append import append_scene
from .. utils.ui import popup_message, init_prefs
from .. utils.system import open_folder
from .. utils.bake import create_bakebasepath, prepare_active, prepare_decals, bake_target_mask, bake_decals_margin_mask, create_decals_mask, bake, apply_decals_mask, split_ao_curv_height_channels, combine, resample, preview, create_empty_map, apply_substance_naming
from .. utils.decal import remove_decal_orphans
from .. utils.object import update_local_view


# TODO: always bake combine masks?


class Bake(bpy.types.Operator):
    bl_idname = "machin3.bake_decals"
    bl_label = "MACHIN3: Bake Decals"
    bl_description = "description"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        if get_prefs().pil and context.mode == 'OBJECT':
            dm = context.scene.DM
            if any([dm.export_bake_color, dm.export_bake_normal, dm.export_bake_aocurvheight, dm.export_bake_masks]):
                return [obj for obj in context.selected_objects if obj.type == 'MESH' and not obj.DM.isdecal and obj.data.uv_layers and not obj.DM.prebakepreviewmats and not obj.name == 'Combined']

    def execute(self, context):
        # only non decal mesh objects with decal children can be baked down to
        targets = [obj for obj in context.selected_objects if obj.type == 'MESH' and not obj.DM.isdecal and obj.data.uv_layers and [child for child in obj.children if child.DM.isdecal] and not obj.DM.prebakepreviewmats and not obj.name == 'Combined']

        if targets:
            start = datetime.now()

            templatepath = os.path.join(get_path(), "resources", "Templates.blend")
            exportpath = os.path.join(get_path(), "assets", "Export")
            bakespath = os.path.join(exportpath, "bakes")
            blendname = os.path.basename(bpy.data.filepath)[:-6] if bpy.data.filepath else ''
            bakebasepath = create_bakebasepath(bakespath, blendname)

            # get the current cyclcles device
            device = context.scene.cycles.device

            dm = context.scene.DM

            width = dm.export_bake_x
            height = dm.export_bake_y
            supersample = int(dm.export_bake_supersample)
            samples = dm.export_bake_samples

            # turn off prefs.edit.use_enter_edit_mode
            init_prefs(context)

            print("\nInfo: Starting to bake decals to %s, at %s with %s samples and %dx Anti Aliasing using the %s --------------------" % (", ".join([target.name for target in targets]), "x".join([str(width), str(height)]), samples, supersample, device))

            if supersample:
                width *= supersample
                height *= supersample

            margin = dm.export_bake_margin
            ray_distance = dm.export_bake_distance
            triangulate = dm.export_bake_triangulate

            combine_bakes = dm.export_bake_combine_bakes and len(targets) > 1
            preview_bakes = dm.export_bake_preview
            open_bake_folder = dm.export_bake_open_folder
            substance_naming = dm.export_bake_substance_naming

            bake_color = dm.export_bake_color
            bake_normal = dm.export_bake_normal
            bake_aocurvheight = dm.export_bake_aocurvheight
            bake_masks = dm.export_bake_masks


            # first, remove any orphan decals, otherwise you may get problems linking decals to the bakescene, especially in local view with orphan decals(not backups)
            remove_decal_orphans(debug=True)

            # get the current(pre bakescene) view_layer
            view_layer = context.view_layer

            # get and set create scene
            bakescene = append_scene(templatepath, "Bake")

            # switch to the bakescene
            context.window.scene = bakescene

            # set the bakescene cycles device
            bakescene.cycles.device = device

            # set the samples
            bakescene.cycles.samples = samples

            # set tilesize according to render size, see https://devtalk.blender.org/t/why-is-texture-baking-so-mind-meltingly-slow/5653
            # based on my tests, I can not see any speed improvement, in fact it's a few seconds slower, tested 256-2048
            # if device == 'CPU':
                # bpy.context.scene.render.tile_x = width
                # bpy.context.scene.render.tile_y = height

            # get baketypes
            baketypes = []

            if bake_color:
                baketypes.append('COLOR')

            if bake_normal:
                baketypes.append('NORMAL')

            if bake_aocurvheight:
                baketypes.append('AO_CURV_HEIGHT')

            if bake_masks:
                baketypes.append('SUBSET')

            # initialize bakes dict, which collects the main bakes per target
            bakes = {}

            for target in targets:
                bakes[target] = {}

            if baketypes:
                # collect support bakes, so they can be removed at the end
                support_bakes = []

                for target in targets:
                    bpy.ops.object.select_all(action='DESELECT')

                    # create and prepare the active, its bake material and the bake image
                    active, bakeimg, bakemat = prepare_active(context, bakescene, target, width, height, triangulate)

                    # """
                    # bake the active's uv mask, with a margin of 0, which is used to create decal bakes with margin onlys outside the UVs
                    target_mask = bake_target_mask(bakescene, bakebasepath, target.name, 'mask', bakeimg, bakemat, margin=0, ray_distance=ray_distance)
                    if bake_masks:
                        bakes[target]['MASKS'] = [target_mask]
                    else:
                        support_bakes.append(target_mask)

                    # bake a combine mask, which is like the target mask, but with margins, and used to combine bakes of multiple targets into one map
                    if combine_bakes:
                        combine_mask = bake_target_mask(bakescene, bakebasepath, target.name, 'combine', bakeimg, bakemat, margin=margin, ray_distance=ray_distance)

                        bakes[target]['COMBINE'] = combine_mask
                        support_bakes.append(combine_mask)

                    normal_decals_mask = None
                    color_decals_mask = None
                    # """

                    for baketype in baketypes:
                        # print("\nInfo: Baking %s to %s from decals:" % (baketype, target.name))

                        # link and prepare the decals for this baketype
                        decals = prepare_decals(context, view_layer, bakescene, target, baketype, debug=False)

                        # """
                        if decals:
                            # bake decals uv masks (margin 0 and margin 3)
                            if baketype in ['NORMAL', 'AO_CURV_HEIGHT', 'SUBSET']:
                                if not normal_decals_mask:
                                    normal_margin_0_mask = bake_decals_margin_mask(bakescene, bakebasepath, target.name, 'normal', bakeimg, decals, margin=0, ray_distance=ray_distance)
                                    normal_margin_3_mask = bake_decals_margin_mask(bakescene, bakebasepath, target.name, 'normal', bakeimg, decals, margin=margin, ray_distance=ray_distance)
                                    support_bakes.extend([normal_margin_0_mask, normal_margin_3_mask])

                                    normal_decals_mask = create_decals_mask(bakebasepath, target.name, 'normal', target_mask, normal_margin_0_mask, normal_margin_3_mask)
                                    if bake_masks:
                                        bakes[target]['MASKS'].append(normal_decals_mask)
                                    else:
                                        support_bakes.append(normal_decals_mask)

                            if baketype in ['COLOR']:
                                if not color_decals_mask:
                                    color_margin_0_mask = bake_decals_margin_mask(bakescene, bakebasepath, target.name, 'color', bakeimg, decals, margin=0, ray_distance=ray_distance)
                                    color_margin_3_mask = bake_decals_margin_mask(bakescene, bakebasepath, target.name, 'color', bakeimg, decals, margin=margin, ray_distance=ray_distance)
                                    support_bakes.extend([color_margin_0_mask, color_margin_3_mask])

                                    color_decals_mask = create_decals_mask(bakebasepath, target.name, 'color', target_mask, color_margin_0_mask, color_margin_3_mask)
                                    if bake_masks:
                                        bakes[target]['MASKS'].append(color_decals_mask)
                                    else:
                                        support_bakes.append(color_decals_mask)

                            # with the decals mask(s) out of the way, bake the main map types
                            bake_map = bake(bakescene, bakebasepath, decals, target.name, baketype, bakeimg, margin=margin, ray_distance=ray_distance)

                            # and collect it
                            bakes[target][baketype] = bake_map

                            # apply the decal mask to the bakes
                            if baketype in ['NORMAL', 'AO_CURV_HEIGHT', 'SUBSET']:
                                apply_decals_mask(normal_decals_mask, bake_map, baketype, target.name)

                                # ao_curv_height maps are support bakes, and will be split into channels later
                                if baketype == 'AO_CURV_HEIGHT':
                                    support_bakes.append(bake_map)

                            elif baketype in ['COLOR']:
                                apply_decals_mask(color_decals_mask, bake_map, baketype, target.name)

                            # unlink the decals again and re-enable paralalx groups, to prepare for the next baketype or target
                            self.unlink_decals(bakescene, decals)

                        # decals matching the bake type don't exist
                        else:

                            # normally nothing needs to be done, but if combine bakes is enabled, it's easiest to just supply empty maps to deal with combining
                            # it's the same for the AO_CURV_HEIGHT map channel split
                            if combine_bakes or baketype == 'AO_CURV_HEIGHT':
                                empty_map = create_empty_map(bakebasepath, baketype, target.name, (width, height))
                                bakes[target][baketype] = empty_map

                                if baketype == 'AO_CURV_HEIGHT':
                                    support_bakes.append(empty_map)
                        # """

                    # """
                    # remove the active and it's material and bake image to prepare for the next target
                    self.remove_active(active, bakemat, bakeimg)

                # remove the bake scene
                bpy.data.scenes.remove(bakescene, do_unlink=True)

                end = datetime.now()
                print("Info: Baking Decals complete, duration: %s ----------------------------------------\n" % (str(end - start)))

                # """
                # """

                # combine the bakes, then add the results as a new baketarget to the baked dict
                if combine_bakes:
                    combine(bakebasepath, bakes, support_bakes)

                # scale down the super sampled bakes
                if supersample:
                    resample(bakes, 1 / supersample)

                # split AO_CURV_HEIGHT maps
                if bake_aocurvheight:
                    split_ao_curv_height_channels(bakebasepath, bakes)

                # remove support bakes
                for path in support_bakes:
                    print("Info: Removing support bake: %s" % (path))
                    os.unlink(path)

                # rename bakes according to substance rules for automatic assignment
                if substance_naming:
                    apply_substance_naming(bakes)

                # preview the main bakes (color, normal, ao, subset)
                if preview_bakes:
                    preview(templatepath, bakes, combine_bakes)

                # inspect the bakes in a file browser
                if open_bake_folder:
                    open_folder(bakebasepath)
                # """

        else:
            popup_message('Select at least one object, that has decal children!', title='Illegal Selection')

        return {'FINISHED'}

    def remove_active(self, active, mat, img):
        bpy.data.meshes.remove(active.data, do_unlink=True)
        bpy.data.materials.remove(mat, do_unlink=True)
        bpy.data.images.remove(img, do_unlink=True)

    def unlink_decals(self, bakescene, decals):
        for decal, decalmats in decals:
            bakescene.collection.objects.unlink(decal)

            # also unmute the parallax again
            for mat, pg in decalmats:
                if pg:
                    pg.mute = False


class RestorePreBakePreviewMaterials(bpy.types.Operator):
    bl_idname = "machin3.restore_pre_bakepreview_materials"
    bl_label = "MACHIN3: Restore Pre-BakePreview Materials"
    bl_description = "Restore Pre-BakePreview Materials"
    bl_options = {'REGISTER', 'UNDO'}

    @classmethod
    def poll(cls, context):
        return context.mode == 'OBJECT' and [obj for obj in context.selected_objects if not obj.DM.isdecal and obj.DM.prebakepreviewmats]

    def execute(self, context):
        remove_images = set()

        for target in [obj for obj in context.selected_objects if not obj.DM.isdecal and obj.DM.prebakepreviewmats]:

            # if the material had no slots at all, a special entry named 'NO_SLOTS' has been created
            if target.DM.prebakepreviewmats[0].name == 'NO_SLOTS':
                target.data.materials.clear()

            else:
                # re-assign the original materials, and remove the preview material and bake images
                for col, slot in zip(target.DM.prebakepreviewmats, target.material_slots):
                    preview_material = slot.material
                    slot.material = col.material

                    # NOTE: collect images for them to be removed at the end
                    # ####: removing them now, has crashed Blender a few times
                    for node in preview_material.node_tree.nodes:
                        if node.type == 'TEX_IMAGE':
                            if node.image:
                                remove_images.add(node.image)

                    bpy.data.materials.remove(preview_material, do_unlink=True)

            # clear the collection prop
            target.DM.prebakepreviewmats.clear()

            # unhide all decals
            decals = [obj for obj in target.children if obj.DM.isdecal and not obj.DM.isbackup]

            for decal in decals:
                decal.hide_set(False)

            # update local view
            update_local_view(context.space_data, [(decal, True) for decal in decals])

        # remove the collected bake images
        for img in remove_images:
            bpy.data.images.remove(img, do_unlink=True)

        return {'FINISHED'}
