import bpy
from bpy.props import BoolProperty
import bmesh
from mathutils import Vector
from .. utils.draw import draw_points


class InitGeneratedUVs(bpy.types.Operator):
    bl_idname = "machin3.init_generated_uvs"
    bl_label = "MACHIN3: Initialize Generated UVs"
    bl_description = "Prepare projected and sliced decals for generated UVs by creating lone verts in the bounding box corners of the decal parent"
    bl_options = {'REGISTER', 'UNDO'}

    evaluated: BoolProperty(name="Use evaluated Geometry", default=False)

    def draw(self, context):
        layout = self.layout

        column = layout.column()
        column.prop(self, "evaluated", toggle=True)

    @classmethod
    def poll(cls, context):
        return [obj for obj in context.selected_objects if obj.type == 'MESH' and obj.DM.isdecal and (obj.DM.isprojected or obj.DM.issliced)]

    def execute(self, context):
        candidates = [obj for obj in context.selected_objects if obj.type == 'MESH' and obj.DM.isdecal and (obj.DM.isprojected or obj.DM.issliced)]

        if candidates:

            init = {}

            for obj in candidates:
                if obj.DM.isprojected:
                    target = obj.DM.projectedon if obj.DM.projectedon else obj.parent if obj.parent else None

                elif obj.DM.issliced:
                    target = obj.DM.slicedon if obj.DM.slicedon else obj.parent if obj.parent else None

                if target:
                    if target in init:
                        init[target].append(obj)

                    else:
                        init[target] = [obj]

            for target, decals in init.items():
                mirrors = []

                # disable mirror mods to get the original mesh bounds, not the evaluated
                if not self.evaluated:
                    mirrors = [mod for mod in target.modifiers if mod.type == 'MIRROR' and mod.show_viewport]

                    for mod in mirrors:
                        mod.show_viewport = False

                context.view_layer.update()

                coords = [Vector(co) for co in target.bound_box]
                # draw_points(coords, mx=target.matrix_world, modal=False)

                # enable the mods again
                for mod in mirrors:
                    mod.show_viewport = True

                # create verts at the bounding box coordinates
                for decal in decals:
                    bm = bmesh.new()
                    bm.from_mesh(decal.data)
                    bm.normal_update()
                    bm.verts.ensure_lookup_table()

                    # first remove any existing loose verts
                    loose_verts = [v for v in bm.verts if not v.link_edges]
                    bmesh.ops.delete(bm, geom=loose_verts, context="VERTS")

                    # then create new ones at the targets bounding box corners
                    for co in coords:
                        bm.verts.new(co)

                    bm.to_mesh(decal.data)
                    bm.clear()

                    decal.data.update()

        return {'FINISHED'}
