import bpy
import os
from . system import makedir
from . mesh import reset_material_indices
from . modifier import add_solidify, add_triangulate
from . material import get_decalmats, get_parallaxgroup_from_decalmat, get_material_output, get_decal_textures, get_decal_texture_nodes, append_material, get_pbrnode_from_mat
from . pil import scale_image
from . object import update_local_view
from .. items import export_baketype_decaltype_mapping_dict


# PATHS

def create_bakebasepath(bakespath, blendname):
    if blendname:
        counter = 0
        folderpath = os.path.join(bakespath, blendname)

        while os.path.exists(folderpath):
            counter += 1
            folderpath = os.path.join(bakespath, "%s_%s" % (blendname, str(counter).zfill(3)))

    else:
        counter = 1
        folderpath = os.path.join(bakespath, '001')

        while os.path.exists(folderpath):
            counter += 1
            folderpath = os.path.join(bakespath, str(counter).zfill(3))

    return makedir(folderpath)


def set_bakepath(bakeimg, bakebasepath, name, baketype):
    bakepath = os.path.join(bakebasepath, "%s_%s.png" % (name, baketype.lower()))

    bakeimg.filepath = bakepath
    return bakepath


# OBJECtS

def prepare_active(context, scene, target, width, height, triangulate):
    """
    duplicate the target and link the duplicate, this is because it will be stripped of its materials and to leave the original untouched
    """

    if triangulate:
        if not any([mod.type == 'TRIANGULATE' for mod in target.modifiers]):
            add_triangulate(target)

    active = target.copy()
    active.data = target.data.copy()

    active.data.materials.clear()
    reset_material_indices(active.data)

    scene.collection.objects.link(active)

    if context.space_data.local_view:
        active.local_view_set(context.space_data, True)

    # select and make active
    active.select_set(True)
    context.view_layer.objects.active = active

    # add a solidify, otherwise the bake goes through the backside, see https://developer.blender.org/T66438
    add_solidify(active, thickness=0.00001)

    # prepare image and material for baking
    img = bpy.data.images.new('BakeImage', width=width, height=height)
    img.file_format = 'PNG'
    img.colorspace_settings.name = 'Non-Color'

    mat = bpy.data.materials.new(name='BakeMat')
    mat.use_nodes = True

    node = mat.node_tree.nodes.new("ShaderNodeTexImage")
    node.select = True
    mat.node_tree.nodes.active = node
    node.image = img

    active.data.materials.append(mat)

    return active, img, mat


def prepare_decals(context, view_layer, bakescene, target, baketype, debug=False):
    """
    return list of (decal, materials) tuples, where materials is itself list of (material, parallax_node_group) tuples
    """
    all_decals = [obj for obj in target.children if obj.DM.isdecal and not obj.DM.isbackup and obj.name in view_layer.objects and obj.visible_get(view_layer=view_layer)]

    decals = []

    # link the decals and select them
    for decal in all_decals:
        decalmats = get_decalmats(decal)

        if decalmats and any([mat.DM.decaltype in export_baketype_decaltype_mapping_dict[baketype] for mat in decalmats]):
            bakescene.collection.objects.link(decal)
            decal.select_set(True)

            # mute any parallax groups, and remember the pg node group, so you can toggle it back at the end
            mats = []

            for mat in decalmats:
                pg = get_parallaxgroup_from_decalmat(mat)

                if pg and not pg.mute:
                    pg.mute = True
                    mats.append((mat, pg))

                else:
                    mats.append((mat, None))

            # clear custom normals on the decal mesh, if found (deactivated for now, as decals with custom normals may be intentional)
            # if decal.data.has_custom_normals:
                # bpy.ops.mesh.customdata_custom_splitnormals_clear({'object': decal})

            decals.append((decal, mats))

    # set local view
    if context.space_data.local_view:
        for obj, _ in decals:
            obj.local_view_set(context.space_data, True)

    if debug:
        for decal, _ in decals:
            print(" » %s" % (decal.name))

    return decals


# SETTINGS

def setup_baking(bakescene, baketype, margin=3, to_active=True, ray_distance=0):
    if baketype == 'NORMAL':
        bakescene.cycles.bake_type = 'NORMAL'
        bakescene.render.bake.normal_b = 'NEG_Z'
    elif baketype == 'EMIT':
        bakescene.cycles.bake_type = 'EMIT'

    bakescene.render.bake.margin = margin
    bakescene.render.bake.use_selected_to_active = to_active
    bakescene.render.bake.cage_extrusion = ray_distance


# BAKING

def bake_target_mask(bakescene, bakebasepath, targetname, bakename, bakeimg, bakemat, margin=3, ray_distance=0):
    setup_baking(bakescene, 'EMIT', margin=margin, to_active=False, ray_distance=ray_distance)
    bakepath = set_bakepath(bakeimg, bakebasepath, targetname, bakename)

    print("Info: Baking %s's %s: %s" % (targetname, bakename, bakepath))

    tree = bakemat.node_tree

    output = get_material_output(bakemat)
    emit = tree.nodes.new('ShaderNodeEmission')
    tree.links.new(emit.outputs['Emission'], output.inputs['Surface'])

    bpy.ops.object.bake(type='EMIT')
    bakeimg.save()
    bakeimg.source = 'GENERATED'

    tree.nodes.remove(emit)

    return bakepath


def bake_decals_margin_mask(bakescene, bakebasepath, targetname, bakename, bakeimg, decals, margin=3, ray_distance=0):
    setup_baking(bakescene, 'EMIT', margin=margin, ray_distance=ray_distance)
    bakepath = set_bakepath(bakeimg, bakebasepath, "%s_%s_%s" % (targetname, bakename, 'margin_%s' % (margin)), 'mask')

    print("Info: Baking %s's %s decals mask, with margin %d: %s" % (targetname, bakename, margin, bakepath))

    decalmats = {mat for _, decalmats in decals for mat, _ in decalmats}

    trees = []

    # create and connect temporary emission shader
    for mat in decalmats:
        tree = mat.node_tree

        texture_nodes = get_decal_texture_nodes(mat)
        alpha = texture_nodes.get('NRM_ALPHA') if bakename == 'normal' else texture_nodes.get('COLOR_ALPHA') if bakename == 'color' else None

        # it's possible alpha can't be fetched, for instance when you have an info decal mat as part of a joined decal, and you are baking normals
        if alpha:
            output = get_material_output(mat)
            lastnode_socket = output.inputs['Surface'].links[0].from_socket

            emit = tree.nodes.new('ShaderNodeEmission')
            tree.links.new(emit.outputs['Emission'], output.inputs['Surface'])
            tree.links.new(alpha.outputs['Alpha'], emit.inputs['Color'])

            trees.append((tree, lastnode_socket, output.inputs['Surface'], emit))

    # bake
    bpy.ops.object.bake(type='EMIT')
    bakeimg.save()
    bakeimg.source = 'GENERATED'

    # recreate the original connections and remove the temporary emission shader
    for tree, socket_out, socket_in, emit in trees:
        tree.nodes.remove(emit)
        tree.links.new(socket_out, socket_in)

    return bakepath


def bake(bakescene, bakebasepath, decals, targetname, baketype, bakeimg, margin=3, ray_distance=0):
    if baketype in ['NORMAL']:
        setup_baking(bakescene, baketype, margin=margin, ray_distance=ray_distance)

    elif baketype in ['COLOR']:
        bakeimg.colorspace_settings.name = 'sRGB'

        setup_baking(bakescene, 'EMIT', margin=margin, ray_distance=ray_distance)

        decalmats = {mat for _, decalmats in decals for mat, _ in decalmats}

        trees = []
        alphas = []

        # create and connect temporary emission shader
        for mat in decalmats:
            tree = mat.node_tree

            texture_nodes = get_decal_texture_nodes(mat)
            color_alpha = texture_nodes.get('COLOR_ALPHA')

            # in joined decals, it's possible this node doesn't exist!
            if color_alpha:
                output = get_material_output(mat)
                lastnode_socket = output.inputs['Surface'].links[0].from_socket

                emit = tree.nodes.new('ShaderNodeEmission')

                tree.links.new(emit.outputs['Emission'], output.inputs['Surface'])
                tree.links.new(color_alpha.outputs['Color'], emit.inputs['Color'])

                trees.append((tree, lastnode_socket, output.inputs['Surface'], emit))

            # temporarily set the alpha mode to NONE to bake the color without alpha applied which is what happens with the sRGB color space + STRAIGHT alpha mode
            textures = get_decal_textures(mat)
            alpha = textures.get('COLOR_ALPHA')

            # in joined decals, it's possible this node doesn't exist!
            if alpha:
                alphas.append(alpha)
                alpha.alpha_mode = 'NONE'

    elif baketype in ['AO_CURV_HEIGHT']:
        setup_baking(bakescene, 'EMIT', margin=margin, ray_distance=ray_distance)

        decalmats = {mat for _, decalmats in decals for mat, _ in decalmats}

        trees = []

        # create and connect temporary emission shader
        for mat in decalmats:
            tree = mat.node_tree

            texture_nodes = get_decal_texture_nodes(mat)
            aocurvheight = texture_nodes.get('AO_CURV_HEIGHT')

            # in joined decals, it's possible this node doesn't exist!
            if aocurvheight:
                output = get_material_output(mat)
                lastnode_socket = output.inputs['Surface'].links[0].from_socket

                emit = tree.nodes.new('ShaderNodeEmission')

                tree.links.new(emit.outputs['Emission'], output.inputs['Surface'])
                tree.links.new(aocurvheight.outputs['Color'], emit.inputs['Color'])

                trees.append((tree, lastnode_socket, output.inputs['Surface'], emit))

    elif baketype in ['SUBSET']:
        setup_baking(bakescene, 'EMIT', margin=margin, ray_distance=ray_distance)

        # decalmats = {mat for _, mat, _ in decals}
        decalmats = {mat for _, decalmats in decals for mat, _ in decalmats}

        trees = []

        # create and connect temporary emission shader
        for mat in decalmats:
            tree = mat.node_tree

            texture_nodes = get_decal_texture_nodes(mat)
            masks = texture_nodes.get('MASKS')

            # in joined decals, it's possible this node doesn't exist!
            if masks:
                output = get_material_output(mat)
                lastnode_socket = output.inputs['Surface'].links[0].from_socket

                emit = tree.nodes.new('ShaderNodeEmission')
                separate = tree.nodes.new('ShaderNodeSeparateRGB')

                tree.links.new(emit.outputs['Emission'], output.inputs['Surface'])
                tree.links.new(separate.outputs['R'], emit.inputs['Color'])
                tree.links.new(masks.outputs['Color'], separate.inputs['Image'])

                trees.append((tree, lastnode_socket, output.inputs['Surface'], emit, separate))

    # bake
    path = set_bakepath(bakeimg, bakebasepath, targetname, baketype)
    print("Info: Baking %s's %s map: %s" % (targetname, baketype.lower(), path))

    if baketype == 'NORMAL':
        bpy.ops.object.bake(type=baketype)
    elif baketype in ['COLOR', 'AO_CURV_HEIGHT', 'SUBSET']:
        bpy.ops.object.bake(type='EMIT')

    # save and reset the bake image
    bakeimg.save()
    bakeimg.source = 'GENERATED'
    bakeimg.colorspace_settings.name = 'Non-Color'

    # reset the alpha mode of color alpha maps again
    if baketype in ['COLOR']:
        for alpha in alphas:
            alpha.alpha_mode = 'STRAIGHT'

    # recreate the original connections and remove the temporary emission shader
    if baketype in ['COLOR', 'AO_CURV_HEIGHT']:
        for tree, socket_out, socket_in, emit in trees:
            tree.nodes.remove(emit)
            tree.links.new(socket_out, socket_in)

    elif baketype == 'SUBSET':
        for tree, socket_out, socket_in, emit, separate in trees:
            tree.nodes.remove(emit)
            tree.nodes.remove(separate)
            tree.links.new(socket_out, socket_in)

    return path


# PROCESSING

def create_empty_map(bakebasepath, baketype, targetname, size):
    from PIL import Image

    if baketype == 'COLOR':
        empty_map = Image.new("RGBA", size, color=(0, 0, 0, 0))

    elif baketype == 'NORMAL':
        empty_map = Image.new("RGB", size, color=(128, 128, 255))

    elif baketype == 'AO_CURV_HEIGHT':
        empty_map = Image.new("RGB", size, color=(255, 128, 128))

    elif baketype == 'SUBSET':
        empty_map = Image.new("L", size, color=0)

    path = os.path.join(bakebasepath, "%s_%s.png" % (targetname, baketype.lower()))
    print("Info: Creating %s's empty %s map: %s" % (targetname, baketype.lower(), path))

    empty_map.save(path)

    return path


def create_decals_mask(bakebasepath, targetname, baketype, target_uv, margin_0_uv, margin_3_uv):
    from PIL import Image, ImageOps

    path = os.path.join(bakebasepath, "%s_%s_mask.png" % (targetname, baketype))

    print("Info: Creating %s's %s mask: %s" % (targetname, baketype, path))

    uv = Image.open(target_uv).convert('L')
    margin0 = Image.open(margin_0_uv).convert('RGBA')
    margin3 = Image.open(margin_3_uv)

    uv_invert = ImageOps.invert(uv)
    margin3.putalpha(uv_invert)
    margin_map = Image.alpha_composite(margin0, margin3).convert('L')

    margin_map.save(path)

    del uv, margin0, margin3, uv_invert, margin_map

    return path


def apply_decals_mask(margin_map, bake_map, baketype, targetname):
    from PIL import Image

    print("Info: Applying decals mask to %s's %s map." % (targetname, baketype.lower()))

    margin = Image.open(margin_map)
    bake_margin = Image.open(bake_map)
    bake_margin.putalpha(margin)

    size = bake_margin.size

    if baketype == 'NORMAL':
        bake_base = Image.new("RGBA", size, color=(128, 128, 255, 255))
    elif baketype == 'COLOR':
        bake_base = Image.new("RGBA", size, color=(0, 0, 0, 0))
    elif baketype == 'AO_CURV_HEIGHT':
        bake_base = Image.new("RGBA", size, color=(255, 128, 128, 255))
    elif baketype == 'SUBSET':
        bake_base = Image.new("RGBA", size, color=(0, 0, 0, 255))

    bake = Image.alpha_composite(bake_base, bake_margin)

    if baketype in ['NORMAL', 'AO_CURV_HEIGHT']:
        bake = bake.convert('RGB')

    elif baketype in ['SUBSET']:
        bake = bake.convert('L')

    # bake.save(bake_map[:-4] + "_" + bake_map[-4:])
    bake.save(bake_map)

    del margin, bake_margin, bake_base, bake


def combine(bakebasepath, bakes, support_bakes):
    from PIL import Image

    # create a dictionary of baketype lists, that allows for easy iteration and combination of bakes
    maps = {}

    for idx, (target, baketypedict) in enumerate(bakes.items()):
        for baketype, path in baketypedict.items():
            if idx == 0:
                maps[baketype] = [path]

            else:
                maps[baketype].append(path)


    # initialize bakes intry for combined bakes
    bakes['COMBINED'] = {}

    # for every baketype except COMBINE, iterate over the bakes and combine them
    for baketype, paths in maps.items():

        # ignore the combine mask and the other suport bake masks
        if baketype in ['COMBINE', 'MASKS']:
            continue

        savepath = os.path.join(bakebasepath, "Combined_%s.png" % (baketype.lower()))
        print("Info: Combining %s bakes: %s" % (baketype.lower(), savepath))

        for idx, path in enumerate(paths):

            # at the first iteration, load the image, it's going to be the base
            if idx == 0:
                combined = Image.open(path).convert('RGBA')

            # for all other iterations, load the image to be combined(the layer), and the combine mask
            else:
                # color maps are already in RGBA and don't need to be masked either
                if baketype == 'COLOR':
                    mask = None
                    layer = Image.open(path)
                else:
                    mask = Image.open(maps['COMBINE'][idx]).convert('L')
                    layer = Image.open(path).convert('RGBA')

                    # mask the layer
                    layer.putalpha(mask)

                # combine it with the previous combined
                combined = Image.alpha_composite(combined, layer)

                del mask, layer

            # at the end, save and add the combined bake to the bakes dict using 'COMBINED' as the target (so it can be resampled and previewed like the others bakes)
            if idx == len(paths) -1:

                if baketype in ['NORMAL', 'AO_CURV_HEIGHT']:
                    combined = combined.convert('RGB')

                elif baketype in ['SUBSET']:
                    combined = combined.convert('L')

                combined.save(savepath)
                del combined

                # append each baketype to the  bakes dict
                bakes['COMBINED'][baketype] = savepath

                # append the combined ao_curv_height map to the support bakes
                if baketype == 'AO_CURV_HEIGHT':
                    support_bakes.append(savepath)


def resample(bakes, factor):
    for target, baketypedict in bakes.items():
        for baketype, path in baketypedict.items():
            # masks have a list of paths
            if baketype == 'MASKS':
                for p in path:
                    print("Info: Resampling bake: %s" % (p))
                    scale_image(p, scale=factor)

            # others have a single one
            else:
                print("Info: Resampling bake: %s" % (path))
                scale_image(path, scale=factor)

    print()


def split_ao_curv_height_channels(bakebasepath, bakes):
    from PIL import Image

    maptype = 'AO_CURV_HEIGHT'

    for target, baketypedict in bakes.items():
        targetname = target.capitalize() if target == 'COMBINED' else target.name

        print("Info: Splitting %s's %s map" % (targetname, maptype.lower()))

        path = baketypedict[maptype]

        aocurvheight_map = Image.open(path)
        channels = aocurvheight_map.split()

        for idx, name in enumerate(maptype.split('_')):
            savepath = os.path.join(bakebasepath, "%s_%s.png" % (targetname, name.lower()))

            print(" » %s's %s map: %s" % (targetname, name, savepath))
            channels[idx].save(savepath)

            bakes[target][name] = savepath

        del aocurvheight_map, channels

    print()


def apply_substance_naming(bakes):
    for target, baketypedict in bakes.items():
        for baketype, path in baketypedict.items():

            if baketype == 'NORMAL':
                newpath = path.replace('_normal.png', '_normal_base.png')
                os.rename(path, newpath)
                bakes[target][baketype] = newpath

            elif baketype == 'AO':
                newpath = path.replace('_ao.png', '_ambient_occlusion.png')
                os.rename(path, newpath)
                bakes[target][baketype] = newpath

            elif baketype == 'CURV':
                newpath = path.replace('_curv.png', '_curvature.png')
                os.rename(path, newpath)
                bakes[target][baketype] = newpath

            elif baketype == 'SUBSET':
                newpath = path.replace('_subset.png', '_id.png')
                os.rename(path, newpath)
                bakes[target][baketype] = newpath

            elif baketype == 'MASKS':
                newpaths = []

                for p in path:
                    if '_color_mask.png' in p:
                        newpath = p.replace('_color_mask.png', '_color_opacity.png')
                        newpaths.append(newpath)

                        os.rename(p, newpath)
                    else:
                        newpaths.append(p)

                bakes[target][baketype] = newpaths


# BAKE PREVIEW

def preview(templatepath, bakes, combine_bakes):
    """
    Preview COLOR, NORMAL, AO and SUBSET bakes
    1. append the BAKE_PREVIEW material
    2. prepare dictionary from the bakes dict, that only contains the bakes we are actually previewing and their target object as the key
    3. collect exsiting target object materials and store them in a collectino prop
    4. assign copies of the BAKE_PREVIEW material
    5. set up the baked textures and unmute the image + related nodes according to the bake types
    """

    def match_pbr_nodes(mat, original_pbrnode, preview_pbrnode):
        for i, pi in zip(original_pbrnode.inputs, preview_pbrnode.inputs):
            pi.default_value = i.default_value

            if pi.name == 'Base Color':
                mixnode = mat.node_tree.nodes.get('Mix')
                mixnode.inputs['Color1'].default_value = pi.default_value

                mixnode = mat.node_tree.nodes.get('Mix.AO')
                mixnode.inputs['Color1'].default_value = pi.default_value

            elif pi.name == 'Roughness':
                valuenode = mat.node_tree.nodes.get('Roughness Value')
                valuenode.outputs['Value'].default_value = pi.default_value

    def set_textures(mat, bakes):
        for baketype, img in bakes.items():
            imgnode = mat.node_tree.nodes.get("BAKE_PREVIEW_%s" % (baketype.lower()))
            imgnode.image = img
            imgnode.mute = False

            if baketype == 'COLOR':
                mat.node_tree.nodes.get('Mix').mute = False
                mat.node_tree.nodes.get('Mix.2').mute = False

            elif baketype == 'AO':
                mat.node_tree.nodes.get('Mix.AO').mute = False
                mat.node_tree.nodes.get('Mix.AO.2').mute = False

                mat.node_tree.nodes.get('Mix.ROUGHNESS').mute = False
                mat.node_tree.nodes.get('Mix.AO').mute = False

                mat.node_tree.nodes.get('Mix.ROUGHNESS.2').mute = False

            elif baketype == 'SUBSET':
                mat.node_tree.nodes.get('Mix Shader').mute = False

    preview_mat = append_material(templatepath, 'BAKE_PREVIEW')

    if preview_mat:

        # collect the maps we want to actually preview in its own dict
        preview_maps = {}

        for target, baketypedict in bakes.items():

            # skip the COMBINED pseudo target, as it's not an object that can carry materials
            if target == 'COMBINED':
                continue

            preview_maps[target] = {}

            for baketype, path in baketypedict.items():

                # we are only interested in COLOR, NORMAL, AO and SUBSET
                # if bakes are combined, fetch the combined maps instead of the individual per target maps
                if baketype in ['COLOR', 'NORMAL', 'AO', 'SUBSET']:
                    preview_maps[target][baketype] = bakes['COMBINED'][baketype] if combine_bakes else path

        # print(preview_maps)

        for target, baketypedict in preview_maps.items():

            # hide the decals, so the bakes can be seen
            decals = [obj for obj in target.children if obj.DM.isdecal and not obj.DM.isbackup]

            for decal in decals:
                decal.hide_set(True)

            # bring in the bakes as images, and set the color space accordingly
            bakes = {}

            for baketype, path in baketypedict.items():
                img = bpy.data.images.load(path)

                if baketype != 'COLOR':
                    img.colorspace_settings.name = 'Non-Color'

                bakes[baketype] = img


            if target.material_slots:

                # collect the current materials per material slot, and assign preview material copies
                for idx, slot in enumerate(target.material_slots):
                    col = target.DM.prebakepreviewmats.add()
                    col.index = idx

                    if slot.material:
                        col.name = slot.material.name
                        col.material = slot.material

                        # set a matched copy of the preview material
                        if slot.material.use_nodes and get_pbrnode_from_mat(slot.material):

                            # get the prb nodes
                            original_pbrnode = get_pbrnode_from_mat(slot.material)

                            preview_material = preview_mat.copy()
                            preview_pbrnode = preview_material.node_tree.nodes.get('Principled BSDF')

                            # set the new material
                            slot.material = preview_material

                            # match the pbr nodes
                            match_pbr_nodes(slot.material, original_pbrnode, preview_pbrnode)

                            # set baked textures
                            set_textures(slot.material, bakes)

                        # set an unmatched copy of the preview material for unsupported materials
                        else:
                            slot.material = preview_mat.copy()
                            set_textures(slot.material, bakes)

                    # set an unmatched copy of the preview material for empty slots
                    else:
                        slot.material = preview_mat.copy()
                        set_textures(slot.material, bakes)

            # set an unmatched copy of the preview material if there aren't any slots, also create a special collection entry to remember this state
            else:
                col = target.DM.prebakepreviewmats.add()
                col.name = 'NO_SLOTS'

                preview_material = preview_mat.copy()

                target.data.materials.append(preview_material)
                set_textures(preview_material, bakes)


        # remove the appended preview mat
        bpy.data.materials.remove(preview_mat, do_unlink=True)
