import bpy
import os
from . material import get_parallaxgroup_from_decalmat, get_decal_texture_nodes, get_decalgroup_from_decalmat
from . registration import get_prefs
from . system import splitpath
from . scene import setup_surface_snapping


def toggle_visibility(state, types, debug=False):
    for type in types:

        # hide
        if state:
            if not type.name.startswith("."):
                if debug:
                    print("Hiding %s." % (type.name))
                type.name = ".%s" % type.name

        # unhide
        else:
            if type.name.startswith("."):
                if debug:
                    print("Unhiding %s." % (type.name))
                type.name = type.name[1:]


def toggle_material_visibility(state, debug=False):
    decalmats = [mat for mat in bpy.data.materials if mat.DM.isdecalmat]

    toggle_visibility(state, decalmats, debug=debug)


def toggle_texture_visibility(state, debug=False):
    decaltextures = [img for img in bpy.data.images if img.DM.isdecaltex]

    toggle_visibility(state, decaltextures, debug=debug)


def toggle_decaltype_collection_visibility(state, debug=False):
    cols = [col for col in bpy.data.collections if col.DM.isdecaltypecol]

    toggle_visibility(state, cols, debug=debug)


def toggle_decalparent_collection_visibility(state, debug=False):
    cols = [col for col in bpy.data.collections if col.DM.isdecalparentcol]

    toggle_visibility(state, cols, debug=debug)


def toggle_glossyrays(state, debug=False):
    decals = [obj for obj in bpy.data.objects if obj.DM.isdecal]

    for decal in decals:
        if debug:
            print(decal)

        decal.cycles_visibility.glossy = state
        decal.data.update()


def toggle_parallax(state, debug=False):
    decalmats = [mat for mat in bpy.data.materials if mat.DM.isdecalmat and mat.DM.decaltype in ["SIMPLE", "SUBSET"]]

    for mat in decalmats:
        if debug:
            print(mat)

        parallaxgroup = get_parallaxgroup_from_decalmat(mat)

        if debug:
            print(parallaxgroup)

        if parallaxgroup:
            parallaxgroup.mute = not state


def toggle_normaltransfer_render(state, debug=False):
    mods = [obj.modifiers.get("NormalTransfer") for obj in bpy.data.objects if obj.DM.isdecal and "NormalTransfer" in obj.modifiers]

    for mod in mods:
        if debug:
            print(mods)

        mod.show_render = state

    bpy.context.evaluated_depsgraph_get()


def toggle_normaltransfer_viewport(state, debug=False):
    mods = [obj.modifiers.get("NormalTransfer") for obj in bpy.data.objects if obj.DM.isdecal and "NormalTransfer" in obj.modifiers]

    for mod in mods:
        if debug:
            print(mods)

        mod.show_viewport = state

    bpy.context.evaluated_depsgraph_get()


def switch_normalinterpolation(state, debug=False):
    decalmats = [mat for mat in bpy.data.materials if mat.DM.isdecalmat and mat.DM.decaltype in ["SIMPLE", "SUBSET", "PANEL"]]

    for mat in decalmats:
        if debug:
            print(mat)

        nrmnode = get_decal_texture_nodes(mat)["NRM_ALPHA"]

        if nrmnode:
            nrmnode.interpolation = state


def switch_colorinterpolation(state, debug=False):
    decalmats = [mat for mat in bpy.data.materials if mat.DM.isdecalmat and mat.DM.decaltype in ["INFO"]]

    for mat in decalmats:
        if debug:
            print(mat)

        colornode = get_decal_texture_nodes(mat)["COLOR_ALPHA"]

        if colornode:
            colornode.interpolation = state


def switch_alpha_blendmode(state, debug=False):
    decalmats = [mat for mat in bpy.data.materials if mat.DM.isdecalmat]

    for mat in decalmats:
        if debug:
            print(mat)

        mat.blend_method = state


def change_ao_strength(state, debug=False):
    decalmats = [mat for mat in bpy.data.materials if mat.DM.isdecalmat and mat.DM.decaltype in ["SIMPLE", "SUBSET", "PANEL"]]

    for mat in decalmats:
        if debug:
            print(mat)

        decalgroup = get_decalgroup_from_decalmat(mat)

        if decalgroup:
            ao = decalgroup.inputs["AO Strength"]

            if ao:
                ao.default_value = state


def invert_infodecals(state, debug=False):
    decalmats = [mat for mat in bpy.data.materials if mat.DM.isdecalmat and mat.DM.decaltype == "INFO"]

    for mat in decalmats:
        if debug:
            print(mat)

        decalgroup = get_decalgroup_from_decalmat(mat)

        if decalgroup:
            invert = decalgroup.inputs["Invert"]

            if invert:
                invert.default_value = int(state)


def switch_edge_highlights(state, debug=False):
    decalmats = [mat for mat in bpy.data.materials if mat.DM.isdecalmat and mat.DM.decaltype in ["SIMPLE", "SUBSET", "PANEL"]]

    for mat in decalmats:
        if debug:
            print(mat)

        decalgroup = get_decalgroup_from_decalmat(mat)

        if decalgroup:
            highlights = decalgroup.inputs["Edge Highlights"]

            if highlights:
                highlights.default_value = float(state)


def toggle_pack_images(state, debug=False):
    """
    1. get all decal textures in the file and collect their users
    2. pack or unpack them depending on the state
    3. remove 0 users images
    """

    textures = [(img, img.users) for img in bpy.data.images if img.DM.isdecaltex]

    if state == 'UNPACKED':
        assetspath = get_prefs().assetspath

    for img, users in textures:
        if users:
            if state == "PACKED":
                if not img.packed_file:
                    print("packing", img.name)
                    img.pack()

            else:
                if img.packed_file:
                    print("unpacking", img.name)

                    # unpack without over writing a file

                    # first try to unpack using 'USE_ORIGINAL' method
                    try:
                        img.unpack(method='USE_ORIGINAL')

                    # the above will fail if the path can not be found, which is a bit odd, considerign no file is written
                    # still, the file will be marked as unpacked succesfully, which is all we need
                    # while at it though, rebuild the path, if the decal/library exists, this will produce usable texture paths
                    # if the decal/library doesn't exist, the unpacked decals will not display properly
                    # TODO: there could be an additional unpack method, which actually writes the image to it's proper location
                    # ####: the problem with that is, it still won't produce usable decals(blend + icon), it will only work for the current scene
                    except:
                        split = splitpath(img.filepath)
                        newpath = os.path.join(assetspath, *split[-3:])

                        img.filepath = newpath

        else:
            print("removing 0 users image", img.name)
            bpy.data.images.remove(img)


def toggle_surface_snapping(state):
    if state:
        setup_surface_snapping(bpy.context.scene)
